# -*- coding: utf-8 -*-

"""
Module latticeadaptor.mappings 
=================================================================

A module containing mappings between elements and attributes for the 
different lattice formats used:
* MadX
* Elegant
* Tracy

"""
